# -*- coding: UTF-8 -*-
# Copyright © 2010-2012 Piotr Ożarowski <piotr@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import logging
from debpython.pydist import parse_pydep, guess_dependency
from debpython.version import DEFAULT, SUPPORTED, debsorted, vrepr, vrange_str

# minimum version required for pycompile/pyclean
MINPYCDEP = 'python:any (>= 2.6.6-7~)'

log = logging.getLogger(__name__)


class Dependencies(object):
    """Store relations (dependencies, etc.) between packages."""

    def __init__(self, package):
        self.package = package
        self.depends = []
        self.recommends = []
        self.suggests = []
        self.enhances = []
        self.breaks = []
        self.rtscripts = []

    def export_to(self, dh):
        """Fill in debhelper's substvars."""
        for i in self.depends:
            dh.addsubstvar(self.package, 'python:Depends', i)
        for i in self.recommends:
            dh.addsubstvar(self.package, 'python:Recommends', i)
        for i in self.suggests:
            dh.addsubstvar(self.package, 'python:Suggests', i)
        for i in self.enhances:
            dh.addsubstvar(self.package, 'python:Enhances', i)
        for i in self.breaks:
            dh.addsubstvar(self.package, 'python:Breaks', i)
        for i in self.rtscripts:
            dh.add_rtupdate(self.package, i)

    def __str__(self):
        return "D=%s; R=%s; S=%s; E=%s, B=%s; RT=%s" % (self.depends, \
                self.recommends, self.suggests, self.enhances, \
                self.breaks, self.rtscripts)

    def depend(self, value):
        if value and value not in self.depends:
            self.depends.append(value)

    def recommend(self, value):
        if value and value not in self.recommends:
            self.recommends.append(value)

    def suggest(self, value):
        if value and value not in self.suggests:
            self.suggests.append(value)

    def enhance(self, value):
        if value and value not in self.enhances:
            self.enhances.append(value)

    def break_(self, value):
        if value and value not in self.breaks:
            self.breaks.append(value)

    def rtscript(self, value):
        if value not in self.rtscripts:
            self.rtscripts.append(value)

    def parse(self, stats, options):
        log.debug('generating dependencies for package %s', self.package)

        pub_vers = sorted(stats['public_vers'].union(stats['ext']))
        if pub_vers:
            dbgpkg = self.package.endswith('-dbg')
            tpl = 'python-dbg' if dbgpkg else 'python'
            minv = pub_vers[0]
            maxv = pub_vers[-1]
            # generating "python2.X | python2.Y | python2.Z" dependencies
            # disabled (see #625740):
            #if dbgpkg:
            #    tpl2 = 'python%d.%d-dbg'
            #else:
            #    tpl2 = 'python%d.%d'
            #self.depend(' | '.join(tpl2 % i for i in debsorted(pub_vers)))

            # additional Depends to block python package transitions
            if minv <= DEFAULT:
                self.depend("%s (>= %d.%d)" % \
                            (tpl, minv[0], minv[1]))
            if maxv >= DEFAULT:
                self.depend("%s (<< %d.%d)" % \
                            (tpl, maxv[0], maxv[1] + 1))

        # make sure pycompile binary is available
        if stats['compile']:
            self.depend(MINPYCDEP)

        for interpreter, version in stats['shebangs']:
            self.depend("%s:any" % interpreter)

        for private_dir, details in stats['private_dirs'].iteritems():
            versions = list(v for i, v in details.get('shebangs', []) if v)

            for v in versions:
                if v in SUPPORTED:
                    self.depend("python%d.%d:any" % v)
                else:
                    log.info('dependency on python%s (from shebang) ignored'
                             ' - it\'s not supported anymore', vrepr(v))
            # /usr/bin/python shebang → add python to Depends
            if any(True for i, v in details.get('shebangs', []) if v is None):
                self.depend('python:any')

            if details.get('compile', False):
                self.depend(MINPYCDEP)
                args = ''
                vr = options.vrange
                if len(versions) == 1:  # only one version from shebang
                    args += "-V %s" % vrepr(versions[0])
                elif vr:
                    # if there are no hardcoded versions in shebang or there
                    # are scripts for different Python versions: compile with
                    # default Python version (or the one requested via X-P-V)
                    args += "-V %s" % vrange_str(vr)
                    if vr == (None, None):
                        pass
                    elif vr[0] == vr[1]:
                        self.depend("python%s:any" % vrepr(vr[0]))
                    else:
                        if vr[0]:  # minimum version specified
                            self.depend("python:any (>= %s)" % vrepr(vr[0]))
                        if vr[1]:  # maximum version specified
                            self.depend("python:any (<< %d.%d)" % \
                                       (vr[1][0], vr[1][1] + 1))

                for pattern in options.regexpr or []:
                    args += " -X '%s'" % pattern.replace("'", r"'\''")
                self.rtscript((private_dir, args))

        if options.guess_deps:
            for fn in stats['requires.txt']:
                # TODO: should options.recommends and options.suggests be
                # removed from requires.txt?
                for i in parse_pydep(fn):
                    self.depend(i)

        # add dependencies from --depends
        for item in options.depends or []:
            self.depend(guess_dependency(item))
        # add dependencies from --recommends
        for item in options.recommends or []:
            self.recommend(guess_dependency(item))
        # add dependencies from --suggests
        for item in options.suggests or []:
            self.suggest(guess_dependency(item))

        log.debug(self)



#! /usr/bin/python
# -*- coding: UTF-8 -*-
# Copyright © 2010 Piotr Ożarowski <piotr@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import os
import sys
from subprocess import Popen, PIPE

skip_sensible_names = True if '--skip-sensible-names' in sys.argv else False
os.chdir(os.path.dirname(__file__))
if os.path.isdir('../debpython'):
    sys.path.append('..')
else:
    sys.path.append('/usr/share/python/debpython/')
from debpython.pydist import sensible_pname


if not os.path.isdir('cache'):
    process = Popen('apt-file -s sources.list -c cache update', shell=True)
    process.communicate()
    if process.returncode != 0:
        sys.stderr.write('Cannot download APT data files')
        exit(1)

# find .egg-info files/directories
process = Popen('apt-file -s sources.list -c cache find -x '
                '"/usr/((share/pyshared)|(lib/python2\.[0-9]/((site)|(dist))-packages)|(share/python-support/[^/]+))/[^/]*\.egg-info"',
                shell=True, stdout=PIPE)
stdout, stderr = process.communicate()
if process.returncode != 0:
    sys.stderr.write('Cannot find packages with Egg metadata')
    exit(2)

processed = set()
result = []
for line in stdout.splitlines():
    pname, path = line.split(': ', 1)
    if pname == 'python-setuptools':
        continue
    egg_name = [i.split('-', 1)[0] for i in path.split('/')\
                if i.endswith('.egg-info')][0]
    if egg_name.endswith('.egg'):
        egg_name = egg_name[:-4]
    if skip_sensible_names and sensible_pname(egg_name) == pname:
        continue
    if egg_name not in processed:
        processed.add(egg_name)
        result.append("%s %s\n" % (egg_name, pname))
        #result.append("%s %s\t%s\n" % (egg_name, pname, path))

result.sort()
fp = open('dist_fallback', 'w')
fp.write('python python\n')
fp.write('setuptools python-pkg-resources\n')
fp.write('wsgiref python (>= 2.5) | python-wsgiref\n')
fp.write('argparse python (>= 2.7) | python-argparse\n')
# wasn't recognized due to .pth file (egg-info is in PIL/ and not in *-packages/)
fp.write('pil python-imaging\n')
fp.writelines(result)



# -*- coding: UTF-8 -*-
# Copyright © 2010-2012 Piotr Ożarowski <piotr@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import logging
import re
from ConfigParser import SafeConfigParser
from os import environ
from os.path import exists, dirname, join
from types import GeneratorType

# will be overriden via debian_defaults file few lines later
SUPPORTED = [(2, 7),]
DEFAULT = (2, 7)

RANGE_PATTERN = r'(-)?(\d\.\d+)(?:(-)(\d\.\d+)?)?'
RANGE_RE = re.compile(RANGE_PATTERN)

log = logging.getLogger(__name__)

# try to read debian_defaults and get a list of supported Python versions and
# the default one from there
_supported = environ.get('DEBPYTHON_SUPPORTED')
_default = environ.get('DEBPYTHON_DEFAULT')
if not _supported or not _default:
    _config = SafeConfigParser()
    _config.read('/usr/share/python/debian_defaults')
    if not _default:
        _default = _config.get('DEFAULT', 'default-version')[6:]
    if not _supported:
        _supported = _config.get('DEFAULT', 'supported-versions')\
                     .replace('python', '')
try:
    DEFAULT = tuple(int(i) for i in _default.split('.'))
except Exception:
    log.exception('cannot read debian_defaults')
try:
    SUPPORTED = tuple(tuple(int(j) for j in i.strip().split('.'))
                            for i in _supported.split(','))
except Exception:
    log.exception('cannot read debian_defaults')


def get_requested_versions(vrange=None, available=None):
    """Return a set of requested and supported Python versions.

    :param available: if set to `True`, return installed versions only,
        if set to `False`, return requested versions that are not installed.
        By default returns all requested versions.
    :type available: bool

    >>> sorted(get_requested_versions([(2, 5), (3, 0)]))
    [(2, 7)]
    >>> sorted(get_requested_versions('')) == sorted(SUPPORTED)
    True
    >>> sorted(get_requested_versions([None, None])) == sorted(SUPPORTED)
    True
    >>> get_requested_versions([(5, 0), None])
    set([])
    """
    if isinstance(vrange, basestring):
        vrange = parse_vrange(vrange)

    if not vrange or list(vrange) == [None, None]:
        versions = set(SUPPORTED)
    else:
        minv = (0, 0) if vrange[0] is None else vrange[0]
        maxv = (99, 99) if vrange[1] is None else vrange[1]
        if minv == maxv:
            versions = set((minv,) if minv in SUPPORTED else tuple())
        else:
            versions = set(v for v in SUPPORTED if minv <= v < maxv)

    if available:
        versions = set(v for v in versions
                       if exists("/usr/bin/python%d.%d" % v))
    elif available is False:
        versions = set(v for v in versions
                       if not exists("/usr/bin/python%d.%d" % v))

    return versions


def parse_vrange(value):
    """Return minimum and maximum Python version from given range.

    >>> parse_vrange('2.4-')
    ((2, 4), None)
    >>> parse_vrange('2.4-2.6')
    ((2, 4), (2, 6))
    >>> parse_vrange('2.4-3.0')
    ((2, 4), (3, 0))
    >>> parse_vrange('-2.7')
    (None, (2, 7))
    >>> parse_vrange('2.5')
    ((2, 5), (2, 5))
    >>> parse_vrange('') == parse_vrange('-') == (None, None)
    True
    """
    if value in ('', '-'):
        return None, None

    match = RANGE_RE.match(value)
    if not match:
        raise ValueError("version range is invalid: %s" % value)
    groups = match.groups()

    if list(groups).count(None) == 3:  # only one version is allowed
        minv = tuple(int(i) for i in groups[1].split('.'))
        return minv, minv

    minv = maxv = None
    if groups[0]:  # maximum version only
        maxv = groups[1]
    else:
        minv = groups[1]
        maxv = groups[3]

    minv = tuple(int(i) for i in minv.split('.')) if minv else None
    maxv = tuple(int(i) for i in maxv.split('.')) if maxv else None

    if maxv and minv and minv > maxv:
        raise ValueError("version range is invalid: %s" % value)

    return minv, maxv


def parse_pycentral_vrange(value):
    """Parse XS-Python-Version.

    >>> parse_pycentral_vrange('current') == (DEFAULT, DEFAULT)
    True
    >>> parse_pycentral_vrange('all')
    (None, None)
    >>> parse_pycentral_vrange('all, >= 2.4')
    ((2, 4), None)
    >>> parse_pycentral_vrange('all, << 3.0')
    (None, (3, 0))
    >>> parse_pycentral_vrange('2.6')
    ((2, 6), (2, 6))
    >>> parse_pycentral_vrange('2.5, 2.6')
    ((2, 5), None)
    >>> parse_pycentral_vrange('>= 2.6.3')
    ((2, 6), None)
    """
    get = lambda x: get_requested_versions(parse_vrange(x))

    current = False
    minv = maxv = None
    hardcoded = set()

    for item in value.split(','):
        item = item.strip()

        if item == 'all':
            continue
        elif item == 'current':
            current = True
            continue

        match = re.match('>=\s*([\d\.]+)', item)
        if match:
            minv = "%.3s" % match.group(1)
            continue
        match = re.match('<<\s*([\d\.]+)', item)
        if match:
            maxv = "%.3s" % match.group(1)
            continue
        match = re.match('^[\d\.]+$', item)
        if match:
            hardcoded.add("%.3s" % match.group(0))

    if len(hardcoded) == 1:
        ver = hardcoded.pop()
        return getver(ver), getver(ver)

    if not minv and hardcoded:
        # yeah, no maxv!
        minv = sorted(hardcoded)[0]

    if current:
        versions = sorted(get("%s-%s" % (minv if minv else '',
                                         maxv if maxv else '')))
        if not versions:
            raise ValueError("version range doesn't match installed Python versions: %s" % value)
        # not really what "current" means...
        if DEFAULT in versions:
            return DEFAULT, DEFAULT
        else:
            return versions[0], versions[0]
    return getver(minv) if minv else None, \
           getver(maxv) if maxv else None


def vrange_str(vrange):
    """Return version range string from given range.

    >>> vrange_str(((2, 4), None))
    '2.4-'
    >>> vrange_str(((2, 4), (2, 6)))
    '2.4-2.6'
    >>> vrange_str(((2, 4), (3, 0)))
    '2.4-3.0'
    >>> vrange_str((None, (2, 7)))
    '-2.7'
    >>> vrange_str(((2, 5), (2, 5)))
    '2.5'
    >>> vrange_str((None, None))
    '-'
    """
    if vrange[0] is vrange[1] is None:
        return '-'
    if vrange[0] == vrange[1]:
        return '.'.join(str(i) for i in vrange[0])
    elif vrange[0] is None:
        return '-' + '.'.join(str(i) for i in vrange[1])
    elif vrange[1] is None:
        return '.'.join(str(i) for i in vrange[0]) + '-'
    else:
        return "%s-%s" % ('.'.join(str(i) for i in vrange[0]),
                          '.'.join(str(i) for i in vrange[1]))


def vrepr(value):
    """
    >>> vrepr(([2, 7], [3, 2]))
    ['2.7', '3.2']
    >>> vrepr(('2.6', '3.1'))
    ['2.6', '3.1']
    >>> vrepr('2.7')
    '2.7'
    >>> vrepr((2, 7))
    '2.7'
    """
    if isinstance(value, basestring):
        return value
    elif not isinstance(value, (GeneratorType, set))\
      and isinstance(value[0], int):
        return '.'.join(str(i) for i in value)

    result = []
    for version in value:
        if isinstance(version, basestring):
            result.append(version)
        else:
            result.append('.'.join(str(i) for i in version))
    return result


def getver(value):
    """Return pair of integers that represent version.

    >>> getver('2.5')
    (2, 5)
    >>> getver('2.6.4')
    (2, 6)
    >>> getver(None)
    ''
    """
    if not value:
        return ''
    return tuple(int(i) for i in value.split('.', 2))[:2]


def debsorted(versions, return_str=None):
    """Return sorted list of versions starting with default Python
    version (if available) then list of suppored versions greater than default
    one followed by reversed list of older versions.

    List of versions sorted this way can be used in Depends field.

    :param vrepr: return string represenatations of versions, by default the
        same format is used as in :param:`versions`

    >>> debsorted([(2, 6), (3, 1), (2, 5), (2, 4), (2, 7)])[0] == DEFAULT
    True
    >>> debsorted(('2.4', '3.2', '2.6', '2.7'))[-1]
    (2, 4)
    >>> debsorted(set([(2, 1), (2, 2)]))
    [(2, 2), (2, 1)]
    >>> debsorted([(2, 1), (2, 2)], return_str=True)
    ['2.2', '2.1']
    """
    result = []
    old_versions = []
    for version in sorted(versions):
        if isinstance(version, basestring):
            version = getver(version)
        if version < DEFAULT:
            old_versions.append(version)
        else:
            result.append(version)
    result.extend(reversed(old_versions))
    if return_str and result:
        return vrepr(result)
    return result



# -*- coding: UTF-8 -*-
# Copyright © 2010-2012 Piotr Ożarowski <piotr@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import with_statement
import codecs
import logging
import os
import re
from cPickle import dumps
from glob import glob
from os.path import exists, isdir, join, split
from shutil import rmtree
from subprocess import PIPE, Popen
from debpython.version import RANGE_PATTERN, getver, get_requested_versions

log = logging.getLogger(__name__)
EGGnPTH_RE = re.compile(r'(.*?)(-py\d\.\d(?:-[^.]*)?)?(\.egg-info|\.pth)$')
SHEBANG_RE = re.compile(r'^#!\s*(.*?/bin/.*?)(python(2\.\d+)?(?:-dbg)?)(?:\s(.*))?[$\r\n].*')
SHAREDLIB_RE = re.compile(r'NEEDED.*libpython(\d\.\d)')
INSTALL_RE = re.compile(r"""
    (?P<pattern>.+?)  # file pattern
    (?:\s+  # optional Python module name:
    (?P<module>[A-Za-z][A-Za-z0-9_.]*)?
    )?
    \s*  # optional version range:
    (?P<vrange>%s)?$
""" % RANGE_PATTERN, re.VERBOSE)
REMOVE_RE = re.compile(r"""
    (?P<pattern>.+?)  # file pattern
    \s*  # optional version range:
    (?P<vrange>%s)?$
""" % RANGE_PATTERN, re.VERBOSE)


def sitedir(version, package=None, gdb=False):
    """Return path to site-packages directory.

    >>> sitedir((2, 5))
    '/usr/lib/python2.5/site-packages/'
    >>> sitedir((2, 7), 'python-foo', True)
    'debian/python-foo/usr/lib/debug/usr/lib/python2.7/dist-packages/'
    """
    if isinstance(version, basestring):
        version = tuple(int(i) for i in version.split('.'))

    if version >= (2, 6):
        path = "/usr/lib/python%d.%d/dist-packages/" % version
    else:
        path = "/usr/lib/python%d.%d/site-packages/" % version

    if gdb:
        path = "/usr/lib/debug%s" % path
    if package:
        path = "debian/%s%s" % (package, path)

    return path


def relpath(target, link):
    """Return relative path.

    >>> relpath('/usr/share/python-foo/foo.py', '/usr/bin/foo', )
    '../share/python-foo/foo.py'
    """
    t = target.split('/')
    l = link.split('/')
    while l[0] == t[0]:
        del l[0], t[0]
    return '/'.join(['..'] * (len(l) - 1) + t)


def relative_symlink(target, link):
    """Create relative symlink."""
    return os.symlink(relpath(target, link), link)


def fix_shebang(fpath, replacement=None):
    """Normalize file's shebang.

    :param replacement: new shebang command (path to interpreter and options)
    """
    try:
        with open(fpath) as fp:
            fcontent = fp.readlines()
        if not fcontent:
            log.debug('fix_shebang: ignoring empty file: %s', fpath)
            return None
    except IOError:
        log.error('cannot open %s', fpath)
        return False

    match = SHEBANG_RE.match(fcontent[0])
    if not match:
        return None
    if not replacement:
        path, interpreter, version, argv = match.groups()
        if path != '/usr/bin':  # f.e. /usr/local/* or */bin/env
            replacement = "/usr/bin/%s" % interpreter
        if interpreter == 'python2':
            replacement = '/usr/bin/python'
        if replacement and argv:
            replacement += " %s" % argv
    if replacement:
        log.info('replacing shebang in %s (%s)', fpath, fcontent[0])
        # do not catch IOError here, the file is zeroed at this stage so it's
        # better to fail dh_python2
        with open(fpath, 'w') as fp:
            fp.write("#! %s\n" % replacement)
            fp.writelines(fcontent[1:])
    return True


def shebang2pyver(fpath):
    """Check file's shebang.

    :rtype: tuple
    :returns: pair of Python interpreter string and Python version
    """
    try:
        with open(fpath) as fp:
            data = fp.read(32)
            match = SHEBANG_RE.match(data)
            if not match:
                return None
            res = match.groups()
            if res[1:3] != (None, None):
                if res[2]:
                    return res[1], getver(res[2])
                return res[1], None
    except IOError:
        log.error('cannot open %s', fpath)


def so2pyver(fpath):
    """Return libpython version file is linked to or None.

    :rtype: tuple
    :returns: Python version
    """

    cmd = "readelf -Wd '%s'" % fpath
    process = Popen(cmd, stdout=PIPE, shell=True)
    match = SHAREDLIB_RE.search(process.stdout.read())
    if match:
        return getver(match.groups()[0])


def clean_egg_name(name):
    """Remove Python version and platform name from Egg files/dirs.

    >>> clean_egg_name('python_pipeline-0.1.3_py3k-py3.1.egg-info')
    'python_pipeline-0.1.3_py3k.egg-info'
    >>> clean_egg_name('Foo-1.2-py2.7-linux-x86_64.egg-info')
    'Foo-1.2.egg-info'
    """
    match = EGGnPTH_RE.match(name)
    if match and match.group(2) is not None:
        return ''.join(match.group(1, 3))
    return name


class memoize(object):
    def __init__(self, func):
        self.func = func
        self.cache = {}

    def __call__(self, *args, **kwargs):
        key = dumps((args, kwargs))
        if key not in self.cache:
            self.cache[key] = self.func(*args, **kwargs)
        return self.cache[key]


def pyinstall(package, vrange):
    """Install local files listed in pkg.pyinstall files as public modules."""
    status = True
    srcfpath = "./debian/%s.pyinstall" % package
    if not exists(srcfpath):
        return status
    versions = get_requested_versions(vrange)

    for line in codecs.open(srcfpath, encoding='utf-8'):
        if not line or line.startswith('#'):
            continue
        details = INSTALL_RE.match(line)
        if not details:
            status = False
            log.warn('%s.pyinstall: unrecognized line: %s',
                     package, line)
            continue
        details = details.groupdict()
        if details['module']:
            details['module'] = details['module'].replace('.', '/')
        myvers = versions & get_requested_versions(details['vrange'])
        if not myvers:
            log.debug('%s.pyinstall: no matching versions for line %s',
                      package, line)
            continue
        files = glob(details['pattern'])
        if not files:
            status = False
            log.error('%s.pyinstall: file not found: %s',
                      package, details['pattern'])
            continue
        for fpath in files:
            fpath = fpath.lstrip('/.')
            if details['module']:
                dstname = join(details['module'], split(fpath)[1])
            elif fpath.startswith('debian/'):
                dstname = fpath[7:]
            else:
                dstname = fpath
            for version in myvers:
                dstfpath = join(sitedir(version, package), dstname)
                dstdir = split(dstfpath)[0]
                if not exists(dstdir):
                    try:
                        os.makedirs(dstdir)
                    except Exception:
                        log.error('cannot create %s directory', dstdir)
                        return False
                if exists(dstfpath):
                    try:
                        os.remove(dstfpath)
                    except Exception:
                        status = False
                        log.error('cannot replace %s file', dstfpath)
                        continue
                try:
                    os.link(fpath, dstfpath)
                except Exception:
                    status = False
                    log.error('cannot copy %s file to %s', fpath, dstdir)
    return status


def pyremove(package, vrange):
    """Remove public modules listed in pkg.pyremove file."""
    status = True
    srcfpath = "./debian/%s.pyremove" % package
    if not exists(srcfpath):
        return status
    versions = get_requested_versions(vrange)

    for line in codecs.open(srcfpath, encoding='utf-8'):
        if not line or line.startswith('#'):
            continue
        details = REMOVE_RE.match(line)
        if not details:
            status = False
            log.warn('%s.pyremove: unrecognized line: %s',
                     package, line)
            continue
        details = details.groupdict()
        myvers = versions & get_requested_versions(details['vrange'])
        if not myvers:
            log.debug('%s.pyremove: no matching versions for line %s',
                      package, line)
            continue
        for version in myvers:
            files = glob(sitedir(version, package) + details['pattern'])
            if not files:
                log.debug('%s.pyremove: nothing to remove: python%d.%d, %s',
                          package, version, details['pattern'])
                continue
            for fpath in files:
                if isdir(fpath):
                    try:
                        rmtree(fpath)
                    except Exception, e:
                        status = False
                        log.error(e)
                else:
                    try:
                        os.remove(fpath)
                    except (IOError, OSError), e:
                        status = False
                        log.error(e)
    return status



# -*- coding: UTF-8 -*-
# Copyright © 2010-2012 Piotr Ożarowski <piotr@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import logging
from os import makedirs, chmod
from os.path import exists, join, dirname

log = logging.getLogger(__name__)


class DebHelper(object):
    """Reinvents the wheel / some dh functionality (Perl is ugly ;-P)"""

    def __init__(self, options):
        self.options = options
        self.packages = {}
        self.python_version = None
        source_section = True
        binary_package = None

        pkgs = options.package
        skip_pkgs = options.no_package

        try:
            fp = open('debian/control', 'r')
        except IOError:
            raise Exception('cannot find debian/control file')

        xspv = xpv = False
        for line in fp:
            if not line.strip():
                source_section = False
                binary_package = None
                continue
            if binary_package:
                if binary_package.startswith('python3'):
                    continue
                if pkgs and binary_package not in pkgs:
                    continue
                if skip_pkgs and binary_package in skip_pkgs:
                    continue
                if line.startswith('Architecture:'):
                    arch = line[13:].strip()
                    # TODO: if arch doesn't match current architecture:
                    #del self.packages[binary_package]
                    self.packages[binary_package]['arch'] = arch
                    continue
            elif line.startswith('Package:'):
                binary_package = line[8:].strip()
                if binary_package.startswith('python3'):
                    log.debug('skipping Python 3.X package: %s', binary_package)
                    continue
                if pkgs and binary_package not in pkgs:
                    continue
                if skip_pkgs and binary_package in skip_pkgs:
                    continue
                self.packages[binary_package] = {'substvars': {},
                                                 'autoscripts': {},
                                                 'rtupdates': [],
                                                 'arch': 'any'}
            elif line.startswith('Source:'):
                self.source_name = line[7:].strip()
            elif source_section:
                if line.lower().startswith('xs-python-version:'):
                    xspv = True
                    if not self.python_version:
                        self.python_version = line[18:].strip()
                if line.lower().startswith('x-python-version:'):
                    xpv = True
                    self.python_version = line[17:].strip()

        if xspv and xpv:
            log.error('Please remove XS-Python-Version from debian/control')

        log.debug('source=%s, binary packages=%s', self.source_name, \
                                                   self.packages.keys())

    def addsubstvar(self, package, name, value):
        """debhelper's addsubstvar"""
        self.packages[package]['substvars'].setdefault(name, []).append(value)

    def autoscript(self, package, when, template, args):
        """debhelper's autoscript"""
        self.packages[package]['autoscripts'].setdefault(when, {})\
                            .setdefault(template, []).append(args)

    def add_rtupdate(self, package, value):
        self.packages[package]['rtupdates'].append(value)

    def save_autoscripts(self):
        for package, settings in self.packages.iteritems():
            autoscripts = settings.get('autoscripts')
            if not autoscripts:
                continue

            for when, templates in autoscripts.iteritems():
                fn = "debian/%s.%s.debhelper" % (package, when)
                if exists(fn):
                    data = open(fn, 'r').read()
                else:
                    data = ''

                new_data = ''
                for tpl_name, args in templates.iteritems():
                    for i in args:
                        # try local one first (useful while testing dh_python2)
                        fpath = join(dirname(__file__), '..',
                                     "autoscripts/%s" % tpl_name)
                        if not exists(fpath):
                            fpath = "/usr/share/debhelper/autoscripts/%s" % tpl_name
                        tpl = open(fpath, 'r').read()
                        if self.options.compile_all and args:
                            # TODO: should args be checked to contain dir name?
                            tpl = tpl.replace('#PACKAGE#', '')
                        else:
                            tpl = tpl.replace('#PACKAGE#', package)
                        tpl = tpl.replace('#ARGS#', i)
                        if tpl not in data and tpl not in new_data:
                            new_data += "\n%s" % tpl
                if new_data:
                    data += "\n# Automatically added by dh_python2:" +\
                            "%s\n# End automatically added section\n" % new_data
                    fp = open(fn, 'w')
                    fp.write(data)
                    fp.close()

    def save_substvars(self):
        for package, settings in self.packages.iteritems():
            substvars = settings.get('substvars')
            if not substvars:
                continue
            fn = "debian/%s.substvars" % package
            if exists(fn):
                data = open(fn, 'r').read()
            else:
                data = ''
            for name, values in substvars.iteritems():
                p = data.find("%s=" % name)
                if p > -1:  # parse the line and remove it from data
                    e = data[p:].find('\n')
                    line = data[p + len("%s=" % name):\
                                p + e if e > -1 else None]
                    items = [i.strip() for i in line.split(',') if i]
                    if e > -1 and data[p + e:].strip():
                        data = "%s\n%s" % (data[:p], data[p + e:])
                    else:
                        data = data[:p]
                else:
                    items = []
                for j in values:
                    if j not in items:
                        items.append(j)
                if items:
                    if data:
                        data += '\n'
                    data += "%s=%s\n" % (name, ', '.join(items))
            data = data.replace('\n\n', '\n')
            if data:
                fp = open(fn, 'w')
                fp.write(data)
                fp.close()

    def save_rtupdate(self):
        for package, settings in self.packages.iteritems():
            pkg_arg = '' if self.options.compile_all else "-p %s" % package
            values = settings.get('rtupdates')
            if not values:
                continue
            d = "debian/%s/usr/share/python/runtime.d" % package
            if not exists(d):
                makedirs(d)
            fn = "%s/%s.rtupdate" % (d, package)
            if exists(fn):
                data = open(fn, 'r').read()
            else:
                data = "#! /bin/sh\nset -e"
            for dname, args in values:
                cmd = 'if [ "$1" = rtupdate ]; then' +\
                      "\n\tpyclean %s %s" % (pkg_arg, dname) +\
                      "\n\tpycompile %s %s %s\nfi" % (pkg_arg, args, dname)
                if cmd not in data:
                    data += "\n%s" % cmd
            if data:
                fp = open(fn, 'w')
                fp.write(data)
                fp.close()
                chmod(fn, 0755)

    def save(self):
        self.save_substvars()
        self.save_autoscripts()
        self.save_rtupdate()



#! /usr/bin/python
# -*- coding: UTF-8 -*-
try:
    from setuptools import setup, find_packages
except ImportError:
    exit()

setup(name='Foo',
      version=0.1,
      description="Foo to rule them all",
      long_description="TODO",
      keywords='foo bar baz',
      author='Piotr Ożarowski',
      author_email='piotr@debian.org',
      url='http://www.debian.org/',
      license='MIT',
      package_dir={'': 'lib'},
      packages=find_packages('lib/'),
      package_data = {'foo': ['jquery.js']},
      zip_safe=False,
      install_requires = ['Mako', 'SQLAlchemy >=0.5', 'Baz [extras]'],
)



#!/usr/bin/env python
from distutils.core import setup, Extension
setup(name="distutils-test",
      version = "0.1",
      author="jbailey",
      author_email="jbailey@debian.org",
      url="http://www.python.org/sigs/distutils-sig/",
      ext_modules=[Extension('foo/bar', ['lib/bar.c'])],
      #py_modules=['package'],
      packages = ["foo"],
      package_dir = {'foo': 'lib'}
     )




# -*- coding: UTF-8 -*-
# Copyright © 2011-2012 Piotr Ożarowski <piotr@debian.org>
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import with_statement
import logging
from os import environ, listdir, remove, rmdir
from os.path import dirname, exists, join, getsize, split
from subprocess import Popen, PIPE

from debpython.pydist import PUBLIC_DIR_RE
from debpython.tools import memoize, sitedir

log = logging.getLogger(__name__)


def parse(fpaths, other=None):
    """Parse namespace_packages.txt files."""
    result = set(other or [])
    for fpath in fpaths:
        with open(fpath, 'r') as fp:
            for line in fp:
                if line:
                    result.add(line.strip())
    return result


@memoize
def load(package=None):
    """Return a set of namespaces to regenerate/clean.

    :param package: limit namespaces to the ones needed by given package
    """
    fpaths = None
    # DESTDIR is used in tests
    nsdir = "%s/usr/share/python/ns/" % environ.get('DESTDIR', '')
    if package:
        # only one package is processed, no need to load all files
        fpath = join(nsdir, package)
        if exists(fpath):
            fpaths = [fpath]
    else:
        # load all files
        if exists(nsdir):
            fpaths = [join(nsdir, i) for i in listdir(nsdir)]

    if fpaths:
        result = set(i.replace('.', '/') for i in parse(fpaths))
    else:
        result = set()
    return result


def add_namespace_files(files, package=None, action=None):
    """Add __init__.py files to given generator."""
    if action is not None:
        namespaces = load(package)
        already_processed = set()
        removal_candidates = set()
    for fn in files:
        yield fn
        if action is None:
            continue
        dpath = dirname(fn)
        if dpath not in already_processed:
            already_processed.add(dpath)
            m = PUBLIC_DIR_RE.match(dpath)
            if m:
                public_dir = m.group()
                while dpath != public_dir:
                    ns_dir = dpath[len(public_dir) + 1:]
                    if ns_dir in namespaces:
                        fpath = join(dpath, '__init__.py')
                        if action is True:
                            try:
                                open(fpath, 'a').close()
                            except Exception:
                                log.error('cannot create %s', fpath)
                            else:
                                yield fpath
                        else:  # action is False
                            # postpone it due to dpkg -S call
                            removal_candidates.add(fpath)
                    already_processed.add(dpath)
                    dpath = split(dpath)[0]

    # now deal with to-be-removed namespace candidates (dpkg -S is expensive)
    # dpgk -S is used just to be safe (in case some other package is providing
    # __init__.py file although it's in /usr/share/python/ns dir)
    if action is False and removal_candidates:
        process = Popen("/usr/bin/dpkg -S %s 2>/dev/null" % \
                        ' '.join(removal_candidates), shell=True, stdout=PIPE)
        # FIXME: len(search_string) > 131072
        stdout, stderr = process.communicate()
        for line in stdout.splitlines():
            ns = line.split(': ', 1)[1]
            if ns in removal_candidates:
                removal_candidates.remove(ns)

        for fpath in removal_candidates:
            try:
                remove(fpath)
            except (IOError, OSError), e:
                log.error('cannot remove %s', fpath)
                log.debug(e)
            else:
                yield fpath


def remove_from_package(package, namespaces, versions):
    """Remove empty __init__.py files for requested namespaces."""
    if not isinstance(namespaces, set):
        namespaces = set(namespaces)
    keep = set()
    for ns in namespaces:
        for version in versions:
            fpath = join(sitedir(version, package), *ns.split('.'))
            fpath = join(fpath, '__init__.py')
            if not exists(fpath):
                continue
            if getsize(fpath) != 0:
                log.warning('file not empty, cannot share %s namespace', ns)
                keep.add(ns)
                break

    # return a set of namespaces that should be handled by pycompile/pyclean
    result = namespaces - keep

    # remove empty __init__.py files, if available
    for ns in result:
        for version in versions:
            dpath = join(sitedir(version, package), *ns.split('.'))
            fpath = join(dpath, '__init__.py')
            if exists(fpath):
                remove(fpath)
                if not listdir(dpath):
                    rmdir(dpath)
        # clean pyshared dir as well
        dpath = join('debian', package, 'usr/share/pyshared', *ns.split('.'))
        fpath = join(dpath, '__init__.py')
        if exists(fpath):
            remove(fpath)
            if not listdir(dpath):
                rmdir(dpath)
    return result
